import numpy as np
cimport numpy as np
cimport cython

cdef extern from "math.h":
    double cos(double theta) nogil
    double acos(double theta) nogil
    double sin(double theta) nogil
    double sqrt(double x) nogil
@cython.boundscheck(True)
@cython.cdivision(True)
@cython.wraparound(False)
@cython.embedsignature(True)

def expmap(np.ndarray theta, projection = False):
    """
    Converts exponential map coordinates to stacked rotation matrices.

    Input:
    theta2d -- Fx3 vector holding exponential map coordinates.
             or simply a 3 vector for one.

    Output:
    R -- Fx3x3 rotation matrices.
    """
    cdef np.ndarray[np.float64_t, ndim=2, mode='c'] theta2d
    cdef double x,y,z
    cdef double x2, y2, z2
    cdef Py_ssize_t F
    cdef Py_ssize_t f
    cdef np.ndarray[np.float64_t, ndim=3] R
    cdef double n, cn, sn
    cdef init_rots = 0
    cdef int i, j, k
    if theta.ndim == 1:
        theta2d = np.atleast_2d(theta)
    else:
        theta2d = theta
    
    
    F = theta2d.shape[0]
    if projection:
        R = np.zeros((F,2,3))
    else:
        R = np.zeros((F,3,3))
    for f in range(F):
        x, y, z = theta2d[f,0], theta2d[f,1], theta2d[f,2]
        
        
        # Calculate normal
        x2, y2, z2 = x * x, y * y, z * z
        n = sqrt(x2 + y2 + z2)
        
        if n == 0:
            R[f,0,0] = R[f,1,1] = 1.0
            if not projection:
                R[f,2,2] = 1.0
        else:
            x, y, z = x/n, y/n, z/n
            x2, y2, z2 = x * x, y * y, z * z
            cn = 1 - cos(n)
            sn = sin(n)
            R[f,0, 0] = 1.0 + cn * (-z2 - y2)
            R[f,0, 1] = -sn * z + cn * x * y
            R[f,0, 2] = sn * y + cn * x * z
            R[f,1, 0] = sn * z + cn * x * y
            R[f,1, 1] = 1.0 + cn * (-z2 - x2)
            R[f,1, 2] = -sn * x + cn * y * z
            
            if not projection:
                R[f,2, 0] = -sn * y + cn * x * z
                R[f,2, 1] = sn * x + cn * y * z
                R[f,2, 2] = 1.0 + cn * (-y2 - x2)
        
    if theta.ndim == 1:
        return R[0]
    else:
        return R
    
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
@cython.embedsignature(True)
def d_expmap(np.ndarray theta, projection = False):
    
    cdef double x,y,z, x2, y2, z2, a,b, c
    cdef Py_ssize_t F, f, i, j
    
    cdef np.ndarray[np.float64_t, ndim=4, mode='c'] dR
    cdef np.ndarray[np.float64_t, ndim=2, mode='c'] theta2d #@DuplicatedSignature
    cdef np.ndarray[np.float64_t, ndim=2, mode='c'] dp = np.zeros((4,3))
    cdef np.ndarray[np.float64_t, ndim=3, mode='c'] J
    cdef double n, sn, cn, ocn, n3
    cdef int stride = 3
    if projection:
        stride = 2
        J = np.zeros((2,3,4))
    else:
        J = np.zeros((3,3,4))
    if theta.ndim == 1:
        theta2d = np.atleast_2d(theta)
    else:
        theta2d = theta
    F = theta2d.shape[0]
    if projection:
        dR = np.zeros((F,2,3,3))
    else:
        dR = np.zeros((F,3,3,3))
    for f in range(F):
        a, b, c = theta2d[f,0], theta2d[f,1], theta2d[f,2]
        
        # Calculate normal
        a2, b2, c2 = a*a, b*b, c*c
        n = sqrt(a2+b2+c2)
        n3 = n*n*n
        if n == 0:
            
            
            dR[f,0,1,2] = -1
            dR[f,0,2,1] = 1
            dR[f,1,0,2] = 1
            dR[f,1,2,0] = -1
            if not projection:
                dR[f,2,0,1] = -1
                dR[f,2,1,0] = 1
        else:
            x, y, z = a/n, b/n, c/n
            x2, y2, z2 = x*x, y*y, z*z
            cn = cos(n)
            ocn = 1 - cn
            sn = sin(n)
            J[ 0, 0, 1] = -2 * ocn * y
            J[ 0, 0, 2] = -2 * ocn * z
            J[ 0, 0, 3] = (-z2 - y2) * sn
        
            J[ 0, 1, 0] = ocn * y
            J[ 0, 1, 1] = ocn * x
            J[ 0, 1, 2] = -sn
            J[ 0, 1, 3] = -cn * z + sn * x * y
        
            J[ 0, 2, 0] = ocn * z
            J[ 0, 2, 1] = sn
            J[ 0, 2, 2] = ocn * x
            J[ 0, 2, 3] = cn * y + sn * x * z
        
            J[ 1, 0, 0] = ocn * y
            J[ 1, 0, 1] = ocn * x
            J[ 1, 0, 2] = sn
            J[ 1, 0, 3] = cn * z + sn * x * y
        
            J[ 1, 1, 0] = -ocn * 2 * x
            J[ 1, 1, 2] = -ocn * 2 * z
            J[ 1, 1, 3] = sn * (-z2 - x2)
        
            J[ 1, 2, 0] = -sn
            J[ 1, 2, 1] = ocn * z
            J[ 1, 2, 2] = ocn * y
            J[ 1, 2, 3] = -cn * x + sn * y * z
            if not projection:
                J[2,0,0] = ocn*z
                J[2,0,1] = -sn
                J[2,0,2] = ocn*x
                J[2,0,3] = -cn*y + sn*x*z
                
                J[2,1,0] = sn
                J[2,1,1] = ocn*z
                J[2,1,2] = ocn*y
                J[2,1,3] = cn*x + sn * y *z
                
                J[2,2,0] = -ocn*2*x
                J[2,2,1] = -ocn*2*y
                J[2,2,3] = sn*(-y2-x2)
            
            # Apply chain rule.
            for i in range(4): # normalized coordinate
                for j in range(3): # unnormalized coordinate
                    if i < 3:
                        dp[i,j] = - theta2d[f,i]*theta2d[f,j]/n3
                        if i==j:
                            dp[i,j] += 1/n
                    else:
                        dp[i,j] = theta2d[f,j]/n
            # Apply chain rule.
            for i in range(4): # normalized coordinate
                for j in range(3): # unnormalized coordinate                        
                    for ii in range(stride): # matrix row
                        for jj in range(3): # matrix col
                            dR[f,ii,jj,j] += J[ii,jj,i] * dp[i,j]
    if theta.ndim == 1:
        return dR[0]
    else:
        return dR
