import numpy as np
from os.path import join
from util import load


class DataSet(object):

    @property
    def n_frames(self):
        return self.tracks.shape[0]
    
    @property
    def n_points(self):
        return self.tracks.shape[2]

    def __init__(self, name, tracks, gt = None, image_collection = None):
        
        self.name = name
        self.image_collection = None
        self.tracks = tracks
        if gt != None:
            self.has_gt = True
            self.gt = gt
        else:
            self.has_gt = False
            self.gt = None
            
        self.image_collection = image_collection if image_collection != None else None 

        # Store counts of frames and points.
        self.num_frames, self.num_points = self.tracks.shape[:2]


    def visualize(self, depth = 100):
        """
        Visualize in a MayaVi Scene.
        
        depth -- Where the image plane and tracks are placed.
        """
        import vis
        v = vis.Visualization()
        
        # Try to show images.
        #if self.impaths:
        if self.image_collection != None:
            imglyph = vis.ImageSequence(self.image_collection, position='center', depth= depth)
            v.add(imglyph, 'Image Plane')
        
        # Add the 2D tracks.
        depths = np.ones((self.tracks.shape[0], 1, self.tracks.shape[2])) * depth
        tracks3d = np.concatenate((self.tracks, depths), axis=1)
        track_glyph = vis.Points3D(tracks3d)
        v.add(track_glyph, '2D Tracks')

        # If there is ground truth, then add that.
        if self.has_gt:
            v.add(vis.Points3D(self.gt, color=(0.0, 0.0, 1.0)), 'Ground Truth')

        return v
    
    def view(self):
        self.visualize().show(blocking=True)

    def __str__(self):
        return """Dataset %s : %d points in %d frames""" % (self.name, self.num_points, self.num_frames)

def load_legacy_dataset(dataset, DATASET_DIR = './datasets'):
    
    # Load the data
    data = load(join(DATASET_DIR, dataset, dataset))
    
    # Assume no images first.
    image_collection = None

    tracks = data.tracks    
    
    if data.has_gt:
        gt = data.gt.copy()
        # Prefer to have jacky look at the camera.
        gt[:,2] *= -1    
    else:
        imdir = join(DATASET_DIR, dataset, 'input')
        from util import skimage_hack
        skimage_hack()
        from skimage.io import ImageCollection
        image_collection = ImageCollection(':'.join([imdir + '/*.' + ending for ending in ['png', 'pgm']]))
        gt = None
    return DataSet(dataset, tracks, gt=gt, image_collection=image_collection)

if __name__ == '__main__':
    from argparse import ArgumentParser
    
    parser = ArgumentParser()
    
    parser.add_argument('dataset')
    
    args = parser.parse_args()
    
    dataset = load_legacy_dataset(args.dataset)
    
    dataset.view()
