import logging
import numpy as np
from traits.api import HasTraits, String, Float, Int, Bool
import datasets
import util
from util import norm, normed
from soup import TriangleSoup

import StringIO
import os
import argparse
from multiprocessing import cpu_count

from util import enable_pyximport
enable_pyximport()
import clrigid

class Parameters(HasTraits):
    """The Paramaters to the local rigidity method."""
    
    # Fixed Random seed (defaulting to a pizza joint's phone number)
    seed = Int(4164390000)
    dsetname = String()
    
    # Performance Parameters.
    use_memory = Bool(False)
    soup_memory = String(os.path.join(os.environ['HOME'], 'tmp', 'memory'))
    num_workers = Int(cpu_count())
    
    # 3SFM Parameters.rec
    lam1 = Float(0.01)
    optimizer = String('lbfgs')

    # Proposal Parameters
    random_pass = Bool(True)
    random_pass_probability = Float(0.25) # Subsamples 25% of tracks for a second delaunay pass.

    # Filtering Parameters.
    rms_outlier_factor = Float(1.5)
    min_angle = Float(10)
    outlier_cutoff = Float(2.5)
    
    # Flip Resolution Parameters.
    sigma_spatial = Float(10)
    def degrees_to_spatial_energy(self, theta):
        return util.geman(theta, self.sigma_spatial)[0]
    
    c_temporal = Float(1.0/50.0)
    def degrees_to_temporal_energy(self, theta):
        return self.c_temporal * theta

    qpboi_iterations = Int(0)
    
    # Reconstruction Parameters.
    dont_form_normal_equations = Bool(False)
    
    # Evaluation Parameters.
    
    # Whether to only allow a single global depth flip when aligning to GT or a per frame flip.
    global_flip = Bool(False)
    
    # Bundle Adjustment Parameters.
    @classmethod
    def populate_argparse(cls, parser):
        traits = cls.class_traits()
        
        # Allow overriding a few things.

        parser.add_argument('--rms-outlier-factor', type=float,
                            default = traits['rms_outlier_factor'].default,
                            help = 'The factor to multiply the median of RMS reprojection errors by to find '
                            'the RMS reprojection error threshold.')
        parser.add_argument('-w', '--num-workers', type=int, default = traits['num_workers'].default,
                            help = 'The number of workers to use.  ' +
                            'This defaults to the number of CPUs available.  ' +
                            'Set this to 0 to use a single process with no workers.')
        parser.add_argument('--seed', type=int, default = traits['seed'].default,
                            help = "Random seed to use for track sampling.  " +
                            "Currently defaulting to a pizza joint's phone number.  " +
                            "Use -1 to allow python to seed itself (likely via the system clock).")
        parser.add_argument('--lam1', type=float, default = traits['lam1'].default,
                            help = 'The factor lambda applied to the squared link length penalty weighting.')
        parser.add_argument('-q', '--qpboi-iterations', type=int, default=traits['qpboi_iterations'].default,
                            help = 'The number of QPBOI iterations to perform to attempt to ' +
                            'improve flip the flip resolution.')
        parser.add_argument('--global-flip', action='store_true',
                            help = 'Use this switch to disable the per-frame flip alignment to ground truth.  ' +
                            'A single flip will be applied to all frames.')
        parser.add_argument('-m', '--use-memory', action='store_true',
                            help='Memorize the results of 3SFM computations to disk for later.')
        parser.add_argument('-d', '--dont-form-normal-equations', action='store_true',
                            help='Avoid forming the normal equations when resolving triangle depths.' + 
                            'This may be slightly more accurate but is a lot slower for big problems.')

    def __init__(self, dsetname, *args, **kwargs):
        super(Parameters, self).__init__(*args, **kwargs)
        self.dsetname = dsetname
        
        if self.use_memory:
            print self.soup_memory
        

class LocallyRigidMotion(object):
    
    @property
    def n_tris(self):
        return len(self.triangles)
    
    @property
    def n_frames(self):
        return self.super_soup.n_frames
    
    @property
    def n_variables(self):
        return self.n_tris * self.n_frames
    
    def __init__(self, dset, params = None):
        
        if params == None:
            params = Parameters()
        assert isinstance(params, Parameters)
        self.dset = dset
        self.params = params
        if self.params.seed == -1:
            self.random = np.random.RandomState()
        else:
            self.random = np.random.RandomState(seed=self.params.seed)
        
    def run(self):
        self.build_soup()
        self.filter_soup()
        self.build_mrf()
        self.resolve_flips()
        self.reconstruct()
        
    #===========================================================================
    # Triangle Soup Building and Filtering.
    #===========================================================================

    def build_soup(self):

        dset = self.dset
        params = self.params
        
        # Build the triangle soup.
        super_soup = TriangleSoup(dset.tracks, params)
        
        n_points = super_soup.n_points
        
        if params.random_pass:
            subset = self.random.permutation(super_soup.n_points)[:int(params.random_pass_probability*n_points)]
        
        # Add triangles to the soup.
        for f in range(super_soup.n_frames):
            super_soup.add_delaunay(f)
            if params.random_pass:
                super_soup.add_delaunay(f, subset=subset)

        # Fit all the triangles.
        super_soup.calculate()
        
        self.super_soup = super_soup
        
    def filter_soup(self):
        
        super_soup = self.super_soup
        params = self.params
        
        # Create the soup we actually work with (rigid tris)
        rms_cutoff = np.median(super_soup.rms) * params.rms_outlier_factor
        idx_rms = np.where(super_soup.rms > rms_cutoff)[0]

        # Calculate statistics for filtering
        fits = super_soup.fit_triangles[0]
        
        # Calculate n_tri x 3 x 3 directional unit vectors.
        d1 = fits[:, :, [1, 2, 0]] - fits
        d1 /= norm(d1, axis=1)[:, np.newaxis]
        d2 = fits[:, :, [2, 0, 1]] - fits
        d2 /= norm(d2, axis=1)[:, np.newaxis]
        
        # This gives the n_tri x 3 angles
        angles = np.degrees(np.arccos((d1 * d2).sum(axis=1)))
        idx_angles = np.where(angles < params.min_angle)[0]
        idx = np.union1d(idx_rms, idx_angles)
        
        # Calculate which triangles have outlier lengths.
        median_length = np.median(super_soup.lengths)
        length_cutoff = median_length * params.outlier_cutoff
        idx_outlier = np.where((super_soup.lengths > length_cutoff).sum(axis=1) > 0)[0]
        idx = np.union1d(idx, idx_outlier)
        
        # Now flip to the opposite.
        idx = np.setdiff1d(np.arange(self.super_soup.n_tris), idx)

        # Filter the soup.
        self.soup = self.super_soup.get_subsoup(idx)
        self.triangles = self.soup.triangles
        
    #===========================================================================
    # Flip Resolution.
    #===========================================================================
    
    def energy(self, flips):
        is_even = flips[self.Is] == flips[self.Js]
        return np.sum(self.energies_even * is_even + self.energies_odd * (1 - is_even))
    def to_idx(self, f, i):
        return self.n_tris * f + i
    def from_idx(self, n):
        return (n / self.n_tris, n % self.n_tris)
    
    def build_mrf(self):
        soup = self.soup
        
        # Calculate norms for everything.        
        logging.info("Calculating Normals.")
        
        # Normalize them all at once.
        P = soup.fit_triangles.copy()
        
        # Calculate Normals.
        norms = np.cross(P[:, :, :, 1] - P[:, :, :, 0], P[:, :, :, 2] - P[:, :, :, 0], axisa=2, axisb=2)
        norms = norms.swapaxes(1, 2)
        norms /= norm(norms, axis=1)[:, np.newaxis, :]

        # Flip every triangle and calculate the flipped norms.
        P[:, :, 2, :] *= -1
        flipnorms = np.cross(P[:, :, :, 1] - P[:, :, :, 0], P[:, :, :, 2] - P[:, :, :, 0], axisa=2, axisb=2)
        flipnorms = flipnorms.swapaxes(1, 2)
        flipnorms /= norm(flipnorms, axis=1)[:, np.newaxis, :]

        # Useful.        
        F = self.n_frames
        n_tris = self.n_tris

        # Create MRF nodes and pairwise terms.
        energies_even = []
        energies_odd = []
        
        # Going to build an odd and even (F-1) X n_tris temporal affinity matrix.
        logging.info("Calculating Temporal Affinities.")
        
        def to_degrees(x, assert_cutoff = 1e-5):
            """Takes values in the interval [-1-assert_cutoff,1+assert_cutoff]
            and clamps to the interval [-1,+1] and finds the arccos of x in degrees.
            """
            assert np.max(x) <= 1 + assert_cutoff
            assert np.min(x) >= -1 - assert_cutoff
            x = np.minimum(1, np.maximum(-1, x))
            return np.degrees(np.arccos(x))
        
        # Compute angles between normals in neighbouring frames.
        e_temporal = np.sum(norms[:F - 1] * norms[1:], axis=1)
        e_temporal = to_degrees(e_temporal)
        o_temporal = np.sum(norms[:F - 1] * flipnorms[1:], axis=1)
        o_temporal = to_degrees(o_temporal)
        
        # Convert to energies.
        e_temporal = self.params.degrees_to_temporal_energy(e_temporal)
        o_temporal = self.params.degrees_to_temporal_energy(o_temporal)
        
        # Set up list of edges in graph.
        Is = (n_tris * np.arange(F - 1)[:, np.newaxis] + np.arange(n_tris)).T.flatten().tolist()
        Js = (n_tris * np.arange(1, F)[:, np.newaxis] + np.arange(n_tris)).T.flatten().tolist()
        energies_even +=  e_temporal.T.flatten().tolist()
        energies_odd += o_temporal.T.flatten().tolist()
        
        logging.info('Calculating Spatial Affinities.')
        fit_triangles = soup.fit_triangles.copy()

        # Get all spatial edges.
        (p, q, i_tri1, i_p1, i_q1, i_tri2, i_p2, i_q2) = soup.get_tris_with_common_edge() #@UnusedVariable
        
        # Find directional vector along edge (p,q)
        d1 = fit_triangles[:, i_tri1, :, i_q1] - fit_triangles[:, i_tri1, :, i_p1]
        d1 = d1.swapaxes(0, 1)
        d1 = normed(d1, axis= -1)

        # And for triangle 2        
        d2 = fit_triangles[:, i_tri2, :, i_q2] - fit_triangles[:, i_tri2, :, i_p2]
        d2 = d2.swapaxes(0, 1)
        d2 = normed(d2, axis= -1)
        
        # Flipped version.
        d2_flip = d2.copy()
        d2_flip[:, :, 2] *= -1
        
        # Take dot products
        even_dot = np.sum(d1 * d2, axis= -1)
        even_dot = np.minimum(1, even_dot)
        odd_dot = np.sum(d1 * d2_flip, axis= -1)
        
        # Find angles.
        even_angles = np.degrees(np.arccos(even_dot))
        odd_angles = np.degrees(np.arccos(odd_dot))
        
        # Convert to energies.
        e_spatial = self.params.degrees_to_spatial_energy(even_angles.T.flatten())
        o_spatial = self.params.degrees_to_spatial_energy(odd_angles.T.flatten())
        
        # Add to list of edges in MRF.
        energies_even += e_spatial.tolist()
        energies_odd += o_spatial.tolist()
        Is += (n_tris * np.arange(F)[:, np.newaxis] + i_tri1).T.flatten().tolist()
        Js += (n_tris * np.arange(F)[:, np.newaxis] + i_tri2).T.flatten().tolist()

        # Store these in the instance.
        logging.info("Creating Pairwise MRF structure")
        self.energies_even = np.asarray(energies_even)
        self.energies_odd = np.asarray(energies_odd)
        self.Is = np.asarray(Is)
        self.Js = np.asarray(Js)
        
    def resolve_flips(self):
        
        logging.info("Running greedy flip resolution.")

        evens = self.energies_even
        odds = self.energies_odd
        weights = np.abs(evens - odds)
        Is, Js = self.Is, self.Js
        
        idx = np.argsort(-weights)
        weights = weights[idx]
        evens = evens[idx]
        odds = odds[idx]
        Is = Is[idx]
        Js = Js[idx]
        flips = np.zeros((self.n_variables,), int)
        
        sets = np.arange(self.n_variables)
        
        # If even is greater, then this should be 1, meaning the triangles prefer the odd flip configuration.
        parities = 1*(evens > odds)
        
        # Call greedy algorithm in Cython.
        clrigid.greedy(Is, Js, parities, flips)
        
        if self.params.qpboi_iterations > 0:
            logging.info("Using QPBOI to improve greedy flip resolution.")
            
            n_edges = len(Is)
            n_nodes = self.n_variables
            
            # This will just crash if its not available.
            import qpbo
            
            # Create the graph with the write number of nodes and edges.
            mrf = qpbo.PyQPBO(n_nodes*2, n_edges*2)
            mrf.AddNode(n_nodes)
            for (I, J, e, o) in zip(Is, Js, evens, odds):
                mrf.AddPairwiseTerm(I, J, e, o, o, e)
    
            # Set the values of these variables.
            for i_variable in range(self.n_variables):
                mrf.SetLabel(i_variable, flips.flat[i_variable])
            # Perform QPBOI iterations.
            for i_qpbo in range(self.params.qpboi_iterations):
                
                mrf.Improve()
                for i_variable in range(self.n_variables):
                    flips.flat[i_variable] = mrf.GetLabel(i_variable)
                    
        # Store flips.  
        self.flips = flips.reshape((self.n_frames, self.n_tris))
    
    #===========================================================================
    # Reconstruction.
    #===========================================================================
    
    def find_reconstructible_segments(self):

        # Define a function that yields edges on the triangle graph.
        def edge_iter():
            for (p1, p2, i_tri1, i_tri1_p1, i_tri1_p2, i_tri2, i_tri2_p1, i_tri2_p2) in self.soup.iter_tris_with_common_edge(): #@UnusedVariable
                yield i_tri1, i_tri2
                
        # Run connected components.
        self.components = util.connected_components(self.n_tris, edge_iter())

        # The points in each component.
        self.component_points = [np.unique(self.triangles[component]) for component in self.components]
        point_to_component_point_maps = []
        for points in self.component_points:
            point_to_component_point_map = np.ones(self.dset.n_points, int) * -1
            for (i_point_component, i_point) in enumerate(points):
                point_to_component_point_map[i_point] = i_point_component
            point_to_component_point_maps.append(point_to_component_point_map)
        self.point_to_component_point_maps = point_to_component_point_maps
            
        #self.components = [c for c in self.components if len(c) > 0]
        logging.info("Found %d resolvable components of sizes %s" % (len(self.components), [len(c) for c in self.components]))

        # Need a mapping from soup triangle indices to component triangle indices
        soup_to_component_maps = []
        for component in self.components:
            soup_to_component_map = np.ones(self.soup.n_tris, int) * -1
            for (i_tri_component, i_tri_soup) in enumerate(component):
                soup_to_component_map[i_tri_soup] = i_tri_component
            soup_to_component_maps.append(soup_to_component_map)
        self.soup_to_component_maps = soup_to_component_maps
        
    def reconstruct(self):
        soup = self.soup
        
        assert isinstance(soup, TriangleSoup)
        
        self.find_reconstructible_segments()
        
        logging.info("Reconstructing frames")
        
        flips = self.flips
        F = soup.num_frames
        
        # Get a copy of the fit triangles and flip them according to flips.
        flipped = soup.fit_triangles.copy()
        fupdown = (1 * (flips[..., np.newaxis] == 0) - 1 * (flips[..., np.newaxis] == 1))
        flipped[:, :, 2, :] *= fupdown
        
        reconstructions = []
        if self.dset.has_gt:
            self.component_rms_rec_errors = []
        
        # Reconstruct each component.
        for (i_component, component) in enumerate(self.components):
            
            logging.info("Reconstructing component %d of size %d" % (i_component, len(component)))

            flipped_component = flipped[:,component,:,:]
            
            # Figure out what points are involved in this component.
            points = self.component_points[i_component]
            
            # Get a subsoup, which allows iteration over triangles adjacent by a point.
            component_soup = self.soup.get_subsoup(component)
            
            # Figure out how many constraints.
            n_constraints = 0
            for point in points:
                point_tris = component_soup.adjacent_tris[point]
                n_point_tris = len(point_tris)
                n_constraints += n_point_tris * (n_point_tris - 1) / 2
            n_constraints += 1
            
            # Build the linear system.
            n_tris = len(component)
            b = np.zeros((n_constraints,F))
            Ab = np.zeros((n_tris, F))
            
            if self.params.dont_form_normal_equations:
                A = np.zeros((n_constraints, n_tris))
            else:
                import scipy.sparse
                A = scipy.sparse.lil_matrix((n_constraints, n_tris))

            # Constraint 0 is that sum of Z values = 0
            #A[0,:] = 1
            for i_tri in range(n_tris): A[0,i_tri] = 1
            
            # Go through points and add equations.
            i_constraint = 1
            for point in points:
                
                # Find triangles intersecting at this point.
                point_tris = component_soup.adjacent_tris[point]
                point_tris = list(point_tris)
                for (i, (i_tri1, pidx1)) in enumerate(point_tris):
                    for (i_tri2, pidx2) in point_tris[i + 1:]:
                        
                        A[i_constraint, i_tri1] = 1
                        A[i_constraint, i_tri2] = -1
                        b[i_constraint,:] = flipped_component[:, i_tri2, 2, pidx2] - flipped_component[:, i_tri1, 2, pidx1]
                        Ab[i_tri1,:] += b[i_constraint,:]
                        Ab[i_tri2,:] -= b[i_constraint,:]
                        i_constraint += 1
            
            # Solve the system with least squares for all frames.
            if self.params.dont_form_normal_equations:
                tri_depths = np.linalg.lstsq(A,b)[0]
            else:
                ATA = np.dot(A.T, A).todense()
                tri_depths = np.linalg.solve(ATA, Ab)
            
            # Set these depths for each triangle in all frames
            flipped[:,component,2,:] += tri_depths.T[:,:,np.newaxis]
    
            # Create an array big enough to hold all the points for the entire dataset.
            P = np.zeros((self.n_frames, 3, self.dset.n_points))
            
            # Average each triangle vertex that corresponds to that point first by adding.
            for i_tri in component:
                tri = self.triangles[i_tri]
                P[:, :, tri] += flipped[:, i_tri]
                
            # Then by dividing out the sums.
            for point in points:
                div = np.sum(component_soup.triangles == point)
                if div != 0:
                    P[:, :, point] /= div

            # Now remove all the points not in this component.
            P = P[:,:,points]

            if self.dset.has_gt:
                logging.info("Aligning to ground truth.")

                def calc_rms_rec_error(x, y, per_frame = False):
                    if per_frame:
                        return util.rms(util.norm(x-y, axis=1), axis=-1)
                    else:
                        return util.rms(util.norm(x-y, axis=1))
                        
                # Get ground truth 3D points.
                P_gt = self.dset.gt[:,:,points]
                
                # Get a Z-flipped version of the reconstruction and the flipped triangles.
                P_flipped = P.copy()
                P_flipped[:,2] *= -1
                flipped_flipped = flipped.copy()
                flipped_flipped[:,component,2,:] *= -1
                
                # Find a good depth translations in both cases.
                dz = (P_gt[:,2].mean(1) - P[:,2].mean(1))
                dz_flipped = (P_gt[:,2].mean(1) - P_flipped[:,2].mean(1))
                
                # Depth translate both.
                P[:,2] += dz[:,np.newaxis]
                P_flipped[:,2] += dz_flipped[:,np.newaxis]
                
                # Depth translate both sets of flipped triangles.
                flipped[:,component,2,:] += dz[:,np.newaxis,np.newaxis]
                flipped_flipped[:,component,2,:] += dz_flipped[:,np.newaxis,np.newaxis]
                
                if self.params.global_flip:
                    
                    # If the flipped is better than use that.
                    if calc_rms_rec_error(P_flipped, P_gt) < calc_rms_rec_error(P, P_gt):
                        P = P_flipped
                        flipped[:,component,...] = flipped_flipped[:,component,...]
                else:
                    # Calculate for which frames the flipped version is better.
                    frame_flips = [calc_rms_rec_error(tmp, P_gt, per_frame=True) for tmp in [P, P_flipped]]
                    frame_flips = np.argmin(np.asarray(frame_flips), axis=0)
                    
                    # Update those frames to the flipped versions.
                    idx = np.nonzero(frame_flips)[0]
                    P[idx,...] = P_flipped[idx,...]
                    for f in idx:
                        flipped[f,component,...] = flipped_flipped[f,component,...]

                rms_rec_error = calc_rms_rec_error(P, P_gt)
                self.component_rms_rec_errors.append(rms_rec_error)
                
                logging.info("Component %d has error %.3f" % (i_component+1, rms_rec_error))
                
            else:
                # Use mean z value of average points to center both the
                # average points and the reconstructed triangle component_soup on
                # the z=0 line.
                z_mu = P[:,2,:].mean(axis=-1)
                P[:,2,:] -= z_mu[:,np.newaxis]
                flipped[:, component, 2, :] -= z_mu[:,np.newaxis,np.newaxis]
    
            reconstructions.append(P)
            
        self.registered_triangles = flipped
        self.reconstructions = reconstructions
        
    #===========================================================================
    # Results and Visualization.
    #===========================================================================
    
    def summary(self):
        
        sum = StringIO.StringIO()
        
        print >>sum, 'Dataset Name:', self.dset.name

        n_points = self.dset.n_points
        
        error_str = 'N/A'
        gt_avail_str = 'YES' if self.dset.has_gt else 'NO'
        print >>sum, 'Ground Truth: ' + gt_avail_str
        
        print >>sum, 'Number of Reconstructed Components %d' % len(self.components)
        for (i_component, component) in enumerate(self.components):
            component_number = i_component + 1
            # Number of points in this component.
            n_points_component = len(self.component_points[i_component])
            
            if self.dset.has_gt:
                error_str = '%.3f' % self.component_rms_rec_errors[i_component]
            
            print >>sum, 'Component %(component_number)d : %(n_points_component)d of %(n_points)d points reconstructed. (RMS 3D Error: %(error_str)s)' % locals()
        
        # Get the string from the IO object and close the object.
        sum_str = sum.getvalue()
        sum.close()
        
        return sum_str
    
    def to_results_dict(self):
        
        results = dict(tracks = self.dset.tracks,
                       has_gt = self.dset.has_gt,
                       triangles = self.soup.triangles, components = self.components,
                        registered_triangles = self.registered_triangles,
                        reconstructions = self.reconstructions,
                        component_points = self.component_points
                        )
        
        if self.dset.has_gt:
            results['gt'] = self.dset.gt
            results['component_rms_rec_errors'] = self.component_rms_rec_errors
        
        return results
    
    def visualize(self):
        v = self.dset.visualize()
        import vis
        for (i_component, component) in enumerate(self.components):
            if len(component) > 1:
                v.add(vis.Triangles(self.registered_triangles[:, component,:,:]), 'Registered Triangles %d' % (i_component+1), visible=i_component == 0)
                tris = self.point_to_component_point_maps[i_component][self.triangles[component]]
                P = self.reconstructions[i_component]
                v.add(vis.TriangleMesh(P, tris, representation='wireframe', color=(0, 0, 1)), 'Reconstruction %d' % (i_component+1), visible=i_component==0)
    
        return v
    
def main():
    
    # Create argument parser that prints out default values.
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description = 'Locally Rigid Motion v0.2 - Jonathan Taylor (jtaylor@cs.toronto.edu)',
        epilog = 'More information is provided in README.txt and the Release Notes PDF at' + 
        ' http://www.cs.toronto.edu/~jtaylor/non-rigid/')
    
    # Which dataset to use.
    parser.add_argument('dataset',
                        help = 'Which dataset to use.  These should be placed in the datasets directory.  ' +
                        'The wind dataset is included as an example, so try that to start.')

    # An option to load a MayaVI visualization immediately afterwards.
    parser.add_argument('-s', '--show', help = "Immediately Show the result.", action='store_true')

    # Verbosity.
    parser.add_argument('-v', '--verbosity', action='count', default=0, help='Increase the verbosity')
   
    # Add the Local Rigidity options.
    Parameters.populate_argparse(parser)

    # Parse the command line arguments.
    args = parser.parse_args()
    
    # Set the logging level based on verbosity.
    level = {
             0:logging.INFO,
             1:logging.DEBUG
    }[args.verbosity]
    logging.basicConfig(level=level)
    logging.getLogger().setLevel(level)

    # Get a set of parameters, passing in the parsed args to 
    # override some defaults.
    params = Parameters(args.dataset, **vars(args))
    
    # Load the dataset.
    dset = datasets.load_legacy_dataset(params.dataset)

    # Create the Local Rigidity Object
    lrm = LocallyRigidMotion(dset, params)

    # Run the algorithm.
    lrm.run()
    
    logging.info('Saving Result')
    os.system('mkdir -p .dat')
    dfile = os.path.join('.dat', args.dataset)
    util.save(dfile, lrm)
    
    os.system('mkdir -p dat')
    dfile = os.path.join('dat', args.dataset)
    util.save(dfile, lrm.to_results_dict())

    if args.show:
        # Show the result.
        lrm.visualize().show()
        
if __name__ == '__main__':
    main()
