import numpy as np
import util
from util import norm
from scipy import optimize
import logging
from expmap import expmap, d_expmap

def linear_trisfm(u, force_valid=True):
    """
    Input:
    u -- Fx2x3 array of orthographic projections of 3 points in F frames
    where u[f,:,i] is the projection of point i in frame f.
    force_valid -- Force 

    Output:
    L -- The recovered 3D lengths.
    e -- residuals of the linear system
    r -- rank of the linear system,
    s -- singular values of the linear system,
    """

    # In the paper, the notation L and l are used for the squared lengths.
    # Here I use the notation M and m instead.
    m = np.sum(np.square(u[:, :, [1, 2, 0]] - u), axis=1)

    def valid_approx():
        return np.sqrt(m).max(axis=0)

    # The matrix in the paper.
    A = -np.ones((3, 3)) + 2 * np.eye(3)

    m0 = m[0]
    m_rest = m[1:]

    # Set up the system of equations.
    coefs = 2 * np.dot((m0 - m_rest), A)
    #consts = np.dot(m0, np.dot(A, m0)) - np.sum(m_rest.T * np.dot(A, m_rest.T), axis=0)
    consts = np.dot(m0, np.dot(A, m0)) - np.sum(np.dot(m_rest, A) * m_rest, axis=1)

    # Solve this system.
    (M, e, r, s) = np.linalg.lstsq(coefs, consts)

    # Convert to lengths
    if np.any(M < 0):
        if force_valid:
            L = valid_approx()
        else:
            raise Exception("3SFM yielded negative squared lengths.")
    else:
        # Otherwise proceed to recover lengths.
        L = np.squeeze(np.sqrt(M))
        
        # Check for valid triangle.
        LS = sorted(L)
        if LS[2] > LS[1] + LS[0]:
            if force_valid:
                L = valid_approx()
            else:
                raise Exception("3SFM yielded an invalid triangle.")

    return (L, e, r, s)

# There are a few different ways to parameterize a triangle model.
# The l=l1,l2,l3 parameterization corresponds to the 3 3D edge lengths between
# points 0-1, 1-2, 2-0
# Alternatively, the g=g1,g2,g3 parameterization is simply a triangle in the
# x-y plane with coordinates (0,0), (g1, 0), (g2,g3)
# Lastly the U-2x3 parameterization holds the x-y coordinates of the
# 3 points in the 3 columns.  Centering the points makes this unique.
 
def l_from_g(g):
    """Recover lengths l from g = x1,x2,y2 paramaterization."""
    x1, x2, y2 = g
    l1 = x1
    l2 = np.sqrt((x2 - x1) ** 2 + y2 ** 2)
    l3 = np.sqrt(x2 ** 2 + y2 ** 2)
    return np.asarray([l1, l2, l3])

def g_from_l(l):
    """Recover g=[x1,x2,y2] paramaterization from lengths l = [l1,l2,l3]."""
    l1, l2, l3 = l
    g = np.zeros((3,))
    g[0] = l1
    g[1] = x2 = (l3 ** 2 - l2 ** 2 + l1 ** 2) / (2.0 * l1)
    y2sq = l3 ** 2 - x2 ** 2
    g[2] = np.sqrt(y2sq)
    return g

def calc_U(gamma):
    """Construct scene points U from euclidian paramaterization gamma.
    """

    U = np.zeros((3, 3))
    U[0, 1] = gamma[0] # x1
    U[0, 2] = gamma[1] # x2
    U[1, 2] = gamma[2]
    
    # Center.
    U[:2] -= U[:2].mean(axis=1)[:, np.newaxis]

    return U

dUdgamma = np.zeros((3, 3, 3))

# Derivative with respect to gamma0.
dUdgamma[0, 1, 0] = 1
dUdgamma[0, :, 0] -= 1 / 3.0

# gamma1
dUdgamma[0, 2, 1] = 1.0
dUdgamma[0, :, 1] -= 1 / 3.0

# gamma2
dUdgamma[1, 2, 2] = 1
dUdgamma[1, :, 2] -= 1 / 3.0

def trisfm_error(x, u, lam1, R0=None):
    """3-SFM Objective Function."""
    U = calc_U(x[-3:])
    
    # Number of frames.
    F = u.shape[0]

    # Get all the projection matrices.
    R = expmap(x[:-3].reshape((F, 3)), projection=True)
    
    if R0 == None:
        RR0 = R
    else:
        # Postmultiply Rotations.
        RR0 = np.einsum('fij,fjk->fik', R, R0)

    # Calculate the projections.
    uhat = np.dot(RR0, U)
    
    # And find the error.
    e = np.sum((uhat - u) ** 2)

    # Calculate lengths.
    l = norm(U[:, [1, 2, 0]] - U, axis=0)

    # Objective is sum squared error plus a penalty on large length squared.
    return e + lam1 * np.sum(l ** 2)

def trisfm_grad(x, u, lam1, R0=None):
    """3-SFM Objective Gradient.."""
    
    g = x[-3:]
    
    U = calc_U(g)

    dUdg = dUdgamma[:2]
    
    F = (len(x) - 3) // 3

    grad = np.zeros((len(x),), float)

    theta = x[:-3].reshape((F, 3))
    R = expmap(theta, projection=True)
    
    if R0 != None:
        R = np.einsum('fij,fjk->fik', R, R0)
    
    uhat = np.dot(R, U)
    res = u - uhat
    dR = d_expmap(theta, projection=True)
    if R0 != None:
        dR = np.einsum('fijp,fjkp->fikp', dR, R0[:, :, :, np.newaxis])

    dRU = np.einsum('fijp,...jkp->fikp', dR, U[ :, :, np.newaxis])
    grad[:-3] = (res[..., np.newaxis] * dRU).sum(axis=1).sum(axis=1).reshape(3 * F)
    
    RdU = np.einsum('fij...,...jkp->fikp', R[..., :2], dUdg)
    grad[-3:] = (res[..., np.newaxis] * RdU).sum(axis=0).sum(axis=0).sum(axis=0)

    grad *= -2
    
    x1, x2, y2 = x[-3:]
    
    # Add terms for length squared penalty.
    grad[-3] += lam1 * (4 * x1 - 2 * x2)
    grad[-2] += lam1 * (4 * x2 - 2 * x1)
    grad[-1] += lam1 * (4 * y2)
    
    return grad

def calc_tri_norms(tris):
    """Should have last two axes 3x3."""
    
    n = np.cross(tris[..., 1] - tris[..., 0], tris[..., 2] - tris[..., 0])
    n /= util.norm(n, axis= -1)[..., np.newaxis]
    return n

def trisfm(u, debug=0, lens=None, callback=None, maxiter=None, lam1=0.01, optimizer='lbfgs', lbfgs_m=30, hist=False):
    """Run the 3-SFM algorithm on th Fx2x3 image projections.
    """
        
    if callback == None:
        callback = lambda: None

    # Make sure we dont change what was passed in.
    u = u.copy()

    F = u.shape[0]

    ubar = u.mean(axis= -1)
    u -= ubar[:, :, np.newaxis]
    
    # Get initial length estimate.
    l0 = linear_trisfm(u, force_valid = True)[0]

    x = np.zeros((3 * F + 3,))
    x[-3:] = g_from_l(l0)
    
    # Init pose to something rotated out of the z-axis.
    x[:-3] = .2
    x[2:-3:3] = 0
    
    # Number of times l is updated.
    U = calc_U(x[-3:])

    # Initialize rotations.
    R0 = np.zeros((F, 3, 3))
    R0[:, 0, 0] = R0[:, 1, 1] = R0[:, 2, 2] = 1

    # Find optimal 2D reflection and rotation.
    R2, t = util.absolute_orientation_2d(U[:2], u[:, :2], allow_reflection=True)
    
    R0[:, :2, :2] = R2
    
    warnflag = 0
    progress = True
    l_hist = []
    err_hist = []
    
    # Keep going as long as we are making progress in the objective.
    while progress:
        progress = False

        # Objective and gradient. 
        def f_error(theta):
            err = trisfm_error(theta, u, lam1, R0=R0)
        
            # Perhaps allow 
            if hist:
                l_hist.append(l_from_g(theta[-3:]))
                err_hist.append(err)
            return err
            
        def f_grad(theta):
            return trisfm_grad(theta, u, lam1, R0=R0)

        # Do a global optimization using LBFGS
        rc = optimize.fmin_l_bfgs_b(
                    f_error,
                    x,
                    m=lbfgs_m,
                    args=(),
                    fprime=f_grad,
                    disp=0)
        x = rc[0]
        bfgsd = rc[2]
        warnflag = bfgsd['warnflag']    
        if bfgsd['warnflag'] != 0:
            logging.warning('BFGS Warning %s' % bfgsd['task'])

        # Update R0
        R0 = np.einsum('fij,fjk->fik', expmap(x[:-3].reshape(F, 3)), R0)
        x[:-3] = 0.0
        
        # Consider updating any frames rotation with any other frames rotation
        U = calc_U(x[-3:])
        R0U = np.dot(R0[:, :2], U)

        # Calculate the error in each frame.
        frame_error = np.square(R0U - u).sum(axis= -1).sum(axis= -1)
        
        for f in range(F):
            
            # Calculate the errors using the rotations from any other frame.
            all_errs = np.square(R0U - u[f]).sum(axis= -1).sum(axis= -1)
            
            # Find the best one.
            f_best = all_errs.argmin()

            # If the best reprojection error is not from this frame, then use
            # that rotation for this frame.
            if f_best != f:
                frame_error[f] = all_errs[f_best]
                R0U[f] = R0U[f_best]
                R0[f] = R0[f_best]
                
                # We made progress so flip this flag, so that we perform another bundle adjustment.
                progress = True
            
    rc = {}
    
    # Calculate the fit triangles.
    fit_triangles = np.dot(R0, U)
    
    # Try to make the flip's consistent.
    norms = calc_tri_norms(fit_triangles)
    
    # Flip so that x component is positive.
    idx = np.nonzero(norms[ :, 2] < 0)[0]
    R0[idx, :, 2] *= -1

    # Recalculate
    fit_triangles = np.dot(R0, U)

    # Translate back to location of projections.
    fit_triangles[:, :2, :] += ubar[:, :, np.newaxis]

    # Calculate error.
    err = trisfm_error(x, u, lam1, R0=R0)
    rms = np.sqrt(trisfm_error(x, u, 0, R0=R0) / float(F) / 3.0)
    
    rc['fit_triangles'] = fit_triangles
    rc['err'] = err
    rc['rms'] = rms
    rc['warnflag'] = warnflag
    rc['x'] = x
    rc['l_hist'] = np.asarray(l_hist)
    rc['err_hist'] = np.asarray(err_hist)
    l = norm(U[:, [1, 2, 0]] - U, axis=0)
    rc['l'] = l
    rc['g'] = x[-3:]
    rc['frame_error'] = frame_error
    return rc
