"""A collection of utility functions."""
import logging
import numpy as np

def enable_pyximport():
    import pyximport
    pyximport.install(setup_args={'include_dirs':[np.get_include()]},
                   reload_support=False)

def skimage_hack():
    """Scikits image currently attempts to configure the root logger
    when it is imported and probably should not.  This hack, simply
    imports skimage and reverts the root logger to its previous state.
    """
    
    # Get a copy of the things skimage tends to change.
    logging._acquireLock()
    try:
        level = logging.root.level
        handlers = logging.root.handlers[:]
    finally:
        logging._releaseLock()
    
    # Import skimage's ImageCollection utility.
    import skimage

    # Reset the two things skimage seems to change.
    logging._acquireLock()
    try:
        logging.root.level = level
        logging.root.handlers = handlers
    finally:
        logging._releaseLock()

def minmax(a, *args, **kwargs):
    return np.min(a, *args, **kwargs), np.max(a, *args, **kwargs)

class Bunch(dict):
    """A dictionary with attribute like access.

    Implementation from:
    http://code.activestate.com/recipes/52308-the-simple-but-handy-collector-of-a-bunch-of-named/
    """

    def __init__(self, **kw):
        dict.__init__(self, kw)
        self.__dict__ = self

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self.update(state)
        self.__dict__ = self

def rms(x, axis=None):
    return np.sqrt(np.mean(np.square(x), axis=axis))

def norm(x, axis= -1):
    return np.sqrt((x * x).sum(axis=axis))

def normed(x, axis = -1):
    if axis < -x.ndim or axis >= x.ndim:
        raise ValueError("axis(=%d) out of bounds" % axis)
    if axis < 0:
        axis += x.ndim
    shape = list(x.shape)
    shape[axis] = 1
    shape = tuple(shape)
    return x / norm(x,axis=axis).reshape(shape)




import pickle as pickle
def load(dfile):
    """Loads a dictionary out of dfile.

    Input:
    dfile -- the name of the file, .dat extension is appended automatically.
    Output:
    The loaded dictionary.
    """
    import sys
    if not dfile.endswith('dat'):
        dfile += '.dat'
    h = open(dfile)
    d = pickle.load(h)
    if isinstance(d, dict):
        d = Bunch(**d)
    h.close()
    return d
    
def save(dfile, d):
    """Saves a dictionary d to dfile.

    Input:
    dfile -- the name of the file, .dat extension is appended automatically.
    d -- the dictionary to save.
    """

    if not dfile.endswith('.dat'):
        dfile += '.dat'
        
    if isinstance(d, Bunch):
        d = dict(**d)
        
    h = open(dfile, 'w')
    pickle.dump(d, h, 2)
    h.close()

def extend_dim(a, axis= -1, extent=1, fill=0.0):
    shape = a.shape
    
    # newshape is really just what we are concatenating.
    newshape = list(a.shape)
    newshape[axis] = extent
    newshape = tuple(newshape)
    other = np.ones(newshape) * fill
    
    return np.concatenate((a, other), axis=axis)
    
import time

class Progress(object):
    
    def __init__(self, n_items= -1, verb='processed', noun='items', min_time = 0):
        self.n_items = n_items
        if n_items == -1:
            raise NotImplementedError() # Must fix progress report for this.
        self.i_item = 0
        self.start_time = time.time()
        self.noun = noun
        self.verb = verb
        self.min_time = min_time
        # The last time a report was made.
        self.last_time = self.start_time
        
    def updated(self):
        
        i_item = self.i_item
        time_since_last = time.time() - self.last_time
        do_report = False
        if time_since_last > self.min_time:
            if not (i_item & (i_item - 1)):   # Power of 2
                do_report = True
        if i_item == self.n_items:    # Last item.
                do_report = True
        if do_report:
            self.report_progress()
    
    def __iadd__(self, x):
        self.i_item += x
        self.updated()
        return self

    def finish(self):
        """Force the fact that the job is finished."""
        self.n_items = self.i_item
        self.updated()
        
    def report_progress(self):
        
        # Calculate rate.
        now = time.time()
        if self.i_item == 0:
            speed = np.nan
        else:    
            speed = (now - self.start_time) / float(self.i_item)
        items_per_second = 1.0 / speed
        
        verb, noun,  = self.verb, self.noun
        i_item, n_items = self.i_item, self.n_items
        if i_item < n_items:
            prog_str = '{i_item} of {n_items} {noun} {verb} ({items_per_second:.2f} per second. ETA: {mins}m{secs:.0f}s)'
            secs = speed * (n_items - i_item)
            mins = int(secs / 60)
            secs -= mins * 60
        else:
            prog_str = '{i_item} of {n_items} {noun} {verb} in {mins}m{secs:.0f}s ({items_per_second:.2f} per second.)'
            secs = now - self.start_time
            mins = int(secs/60)
            secs -= mins * 60
        
        logging.info(prog_str.format(**locals()))
        
        # Update the last time reported.
        self.last_time = now
        
    def __str__(self):
        verb, noun = self.verb, self.noun
        i_item, n_items = self.i_item, self.n_items
        return "Progress object : {i_item} of {n_items} {noun} {verb}".format(**locals())


def connected_components(n_nodes, edge_iter):
    """
    Finds the connected components of a graph with n_nodes nodes
    
    Input:
    n_nodes -- The number of nodes in the network.
    edge_iter -- An iterator that yields (i_node1, i_node2) edge pairs.
    
    Returns:
    components -- list of array of int indices, specifying the components.
    """
    
    # Assign each node it's own component id.
    ids = np.arange(n_nodes)

    i_edges = 0    
    # Go through edges.
    for (i_node1, i_node2) in edge_iter:

        # Check if they are not in the same component.
        if ids[i_node1] != ids[i_node2]:
            
            # Merge these two components by labelling all nodes
            # with ids[i_node2] to ids[i_node1].
            ids[np.nonzero(ids == ids[i_node2])[0]] = ids[i_node1]
        
        i_edges += 1

    # Find all the components.
    components = [np.nonzero(ids == component_id)[0] for component_id in np.unique(ids)]
    
    # Sort by decreasing size.
    components = sorted(components, key=len, reverse = True)

    return components

def geman(e, sigma):
    sigma = float(sigma)
    sigma2 = sigma * sigma
    
    e2 = np.square(e)
    sigma2_plus_e2 = sigma2 + e2
    
    return e2 / sigma2_plus_e2, 2 * e * sigma2 / (sigma2_plus_e2 * sigma2_plus_e2)

def huber(e, delta):
    """
    http://en.wikipedia.org/wiki/Huber_loss_function
    """
    delta = float(delta)
    e_abs = np.abs(e)
    I_le_delta = 1.0*(e_abs <= delta)
    
    e2 = np.square(e)
    
    
    rho = (.5 * e2) * I_le_delta +  delta * (e_abs - delta/2.0) * (1-I_le_delta)
    
    d_rho = e * I_le_delta + delta*np.sign(e) * (1-I_le_delta)
    
    return rho, d_rho

def rotation_2d(theta):
    """Calculate 2D rotation matrices from angles theta.
    
    R = [[cos(theta), -sin(theta)],
        [sin(theta), cos(theta)]]
    
    Will broadcast across theta to give you a theta.shapex2x2 array."""
    
    # In case it was just a python numeric passed in.
    theta = np.asarray(theta)
    
    # Figure out the output shape.
    shape = tuple(list(theta.shape) + [2, 2])
    
    # Should be this size.
    R = np.zeros(shape)
    
    # Fill in.
    c = np.cos(theta)
    s = np.sin(theta)
    R[..., 0, 0] = c
    R[..., 0, 1] = -s
    R[..., 1, 0] = s
    R[..., 1, 1] = c
    
    return R


def absolute_orientation_2d(A, B, allow_reflection=False):
    """
    Find t, R to minimize
    
    \sum_{i=1}^N ||b_i - (R(theta)a_i + t)||^2
    
    returns t, theta for each instance (as this can broadcast say to F instances).
    From "On the Absolute Orientation Problem in Computer Vision" by Rodney Roberts, Daniel Repperger, 2009.

    """
    assert A.shape[-2] == B.shape[-2] == 2, "Only works for 2D points"
    assert A.shape[-1] == B.shape[-1], "Must have same number of points"
    
    if allow_reflection:
        
        R, t = absolute_orientation_2d(A, B)
        
        # Get a reflected copy.
        A_reflected = A.copy()
        A_reflected[..., 1, :] *= -1
        
        # Solve for the reflected point set.
        R_reflected, t_reflected = absolute_orientation_2d(A_reflected, B)
        R_reflected[..., :, 1] *= -1

        # Decide which is best.
        B_hat1 = np.einsum('...ij,...jk->...ik', R, A) + t[..., :, np.newaxis]
        B_hat2 = np.einsum('...ij,...jk->...ik', R_reflected, A) + t_reflected[..., :, np.newaxis]

        # Calculate the error under optimal rotation and reflection.
        N = B.shape[-1]
        err1 = np.square(B - B_hat1).reshape(-1, 2, N).sum(axis= -1).sum(axis= -1)
        err2 = np.square(B - B_hat2).reshape(-1, 2, N).sum(axis= -1).sum(axis= -1)
        
        # Choose the right rotations and translations.
        use_reflect = np.argmin(np.asarray([err1, err2]), axis=0)
        use_reflect = np.nonzero(use_reflect)[0]
        R[use_reflect] = R_reflected[use_reflect]
        t[use_reflect] = t_reflected[use_reflect]
        
        return R, t
    
    # Calculate mean of both.
    A_bar = A.mean(axis= -1)
    B_bar = B.mean(axis= -1)

    # Center both
    A = A - A_bar[..., np.newaxis]
    B = B - B_bar[..., np.newaxis]
    
    # Put into complex coordinates.
    A = A[..., 0, :] - 1j * A[..., 1, :] # Conjugate.
    B = B[..., 0, :] + 1j * B[..., 1, :]
    
    # Calculate the angle.
    theta = np.angle((A * B).sum(axis= -1))
    
    # Build rotation matrices
    R = rotation_2d(theta)    
    
    # Calculate translation.
    t = B_bar - np.einsum('...ij,...j->...i', R, A_bar)
    
    return R, t

